#!/usr/local/bin/ruby

require 'wide'

def errp(s)
  $stderr.print s.inspect, "\n"
end

module Escaper
  ESC = /\e/
  def esc_to_space; gsub! ESC, ' '  end
  def esc_to_enter; gsub! ESC, "\n" end
  def esc_to_empty; gsub! ESC, ''   end
end

module Rbml
  def rbml_encode(s)
    s.gsub!(/%%/, "\a")
    s.gsub!(/%(\w+?):(.*?)%/) {
      act = $1
      src = $2
      src.extend Escaper
      src.gsub!(/\a/, '%')
      errp act if $DEBUG
      errp src if $DEBUG
      case act
      when 'e'
        src.esc_to_enter
        Html::html_decode src
        begin
          e = eval src.to_s
        rescue
          e = $!
        end
        if e then e.to_s else '[miss eval]' end
      when 'c'
        src.esc_to_space
        %!<code>#{src}</code>!
      when 'v'
        src.esc_to_space
        %!<var>#{src}</var>!
      when 's'
        src.esc_to_space
        %!<strong>#{src}</strong>!
      when 'html'
        src.esc_to_space
        Html::html_decode src
        src.to_s
      when 'name'
        src.strip!
        src.esc_to_empty
        /([^!]+)(!)?(.*)/ =~ src
        if $2 then a, r = $1, $3 else a, r = $1, $1 end
        %!<a name="#{a}"><br>#{r}</a>!
      when 'href', 'a'
        src.strip!
        src.esc_to_empty
        /([^!]+)(!)?(.*)/ =~ src
        if $2 then a, r = $1, $3 else a, r = $1, $1 end
        %!<a href="#{a}">#{r}</a>!
      when 'mref', 'm'
        src.strip!
        src.esc_to_empty
        a = ''
        /([A-Z]\w*)?(#|\.|::)?(.*)/ =~ src
        if $1
          class_ident = $1
          method_name = $3
          a = class_ident + '.html'
          m = parse_method_name method_name
          if m.nil? then src = class_ident else a << '#' << m end
        else
          m = parse_method_name src
          if m.nil? then ; else a << '#' << m end
        end
        %!<a href="#{a}"><code>#{src}</code></a>!
      when 'method'
        # method anchor
        src.esc_to_empty
        src = src.strip.chomp
        m = parse_method_name src
        if m.nil?
          $stderr.print "error: method name is nil: #{src}\n"
        end
        %!<a name="#{m}"></a><br><h3 class="method">#{src}</h3>!
      when 'include'
        ret = ''
        IO.foreach(src.strip) { |i| ret << i }
        ret
      when 'nil'
        %!!
      else
        %![miss #{act}]!
      end
    }
    s.gsub!(/\a/, '%')
    s
  end
  module_function :rbml_encode

  def here_decode(src, act)
    src.gsub!(/\e%%/, "\e%")
    case act
    when 'e'
      src.esc_to_enter
      begin
        s = eval src.to_s
      rescue
        s = $!
      end
      if s then s.to_s else '<code> [eval: nil] </code>' end
    when 'r'
      src.esc_to_enter
      Html::html_encode src
      %!<pre>#{src}</pre>!
    when 'html'
      src.esc_to_enter
      src.to_s
    when 'ruby'
      src.esc_to_enter
      Html::html_encode src
      %!<pre class="rb">#{src}</pre>!
    when 'c'
      src.esc_to_enter
      Html::html_encode src
      %!<pre class="c">#{src}</pre>!
    when 'item'
      s = src.to_s
      rbml_encode s

      r = s.split(/\e/)
      r.shift

      a = [nil]
      a.push '<ul>'
      pre_lev = 0

      r.each do |i|
        /^ */ =~ i
        indent, string = $&, $'
        lev = indent.length / 2
        1.upto(lev - pre_lev) { a.push %!#{indent}<ul>!  } if pre_lev < lev
        1.upto(pre_lev - lev) { a.push %!#{indent}</ul>! } if pre_lev > lev
        a.push %!#{indent}<li>#{string}!
        pre_lev = lev
      end

      0.upto(pre_lev) { |i| a.push %!#{'  '*(pre_lev-i)}</ul>! }
      a.join "\n"
    when 'enum'
      s = src.to_s
      rbml_encode s

      r = s.split(/\e/)
      r.shift

      a = [nil]
      a.push '<ol>'
      pre_lev = 0

      r.each do |i|
        /^ */ =~ i
        indent, string = $&, $'
        lev = indent.length / 2
        1.upto(lev - pre_lev) { a.push %!#{indent}<ol>!  } if pre_lev < lev
        1.upto(pre_lev - lev) { a.push %!#{indent}</ol>! } if pre_lev > lev
        a.push %!#{indent}<li>#{string}!
        pre_lev = lev
      end

      0.upto(pre_lev) { |i| a.push %!#{'  '*(pre_lev-i)}</ol>! }
      a.join "\n"
    when 'table'
      s = src.to_s
      rbml_encode s
      spec, ceps = 'th', 'td'
      a = [nil]
      a.push '<table border>'
      r = s.split(/\e/)
      r.shift
      r.each { |i| /^ */ =~ i
        indent, string = $&, $'
        if string =~ /--/
          spec, ceps = ceps, spec
          next
        end
        a.push indent + '<tr>'
        string.split(/\|/).each { |i|
          a.push indent + %!  <#{spec}>#{i.strip}</#{spec}>!
        }
        a.push indent + '</tr>'
      }
      a.push '</table>'
      a.join "\n"
    else
      %![miss #{act}]!
    end
  end
  module_function :here_decode

  def parse_method_name(src)
    src =~ /^(\w+)(.*)/
    name = $1
    rest = $2
    if name.nil? then return nil end
    case name
    when 'self'
      case rest[0]
      when ?[
        rest =~ /^(\[.*?\])(.*)/ ; name << '_aref' ; rest = $2
      when ?<
        case rest[1]
        when ?=
          case rest[2]
          when ?> ; name << '_cmp' ; rest = rest[3...-1]
          else      name << '_le'  ; rest = rest[2...-1]
          end
        else name << '_lt' ; rest = rest[1...-1]
        end
      when ?>
        case rest[1]
        when ?= ; name << '_ge' ; rest = rest[2...-1]
        else      name << '_gt' ; rest = rest[1...-1]
        end
      when ?=
        case rest[1]
        when ?=
          case rest[2]
          when ?= ; name << '_eqq' ; rest = rest[3...-1]
          else      name << '_eq'  ; rest = rest[2...-1]
          end
        when ?~ ; name << '_match'  ; rest = rest[2...-1]
        else      name << '_assign' ; rest = rest[1...-1]
        end
      end
    else
      case rest[0]
      when ?? ; name << '_p'    ; rest = rest[1...-1]
      when ?! ; name << '_bang' ; rest = rest[1...-1]
      end
    end
    name
  end
  module_function :parse_method_name

end

module Html
  def html_encode(src)
    src.gsub!(/&/, '&amp;')
    src.gsub!(/</, '&lt;')
    src.gsub!(/>/, '&gt;')
  end
  module_function :html_encode

  def html_decode(src)
    src.gsub!(/&lt;/i, '<')
    src.gsub!(/&gt;/i, '>')
    src.gsub!(/&amp;/i, '&')
  end
  module_function :html_decode

  def html_anchor(src, cap)
    %!<a href="#{src}.html">#{cap}</a>!
  end
  module_function :html_anchor

  module Plain
    include Escaper

    def initialize(*args)
      super
      @act = nil
    end
    attr :act, true

    def markup
      Rbml::rbml_encode self if @act.nil?
      gsub!(/-\e/, '')
      gsub!(/\e/, ' ')
      gsub!(/\f/, '<br><br>')
      display
    end
  end

  class Dl
    class Item
      def initialize(node,spec)
        @node = node
        @spec = spec
      end
      attr :node
      attr :spec

      def markup
        print '<',@spec,'>'
        @node.markup
        print "\n"
      end
    end

    def initialize(nest)
      @nest = nest
      @item = []
    end

    def dt_push(node)
      @item.push Item.new node, 'dt'
    end

    def dd_push(node)
      @item.push Item.new node, 'dd'
    end

    def push(item)
      @item.push item
    end

    def last
      @item[-1]
    end

    def markup
      # for lynx
      while last.is_a?(Item) && last.spec == 'dd' && last.node.to_s !~ /\w/
        @item.pop
      end

      print '  ' * @nest,  '<dl>', "\n"
      @item.each { |i|
        print '<dd>', "\n" if i.is_a?(Dl)  # for IE
        print '  ' * @nest
        print '  ' if i.is_a?(Item) && i.spec == 'dd'
        i.markup
      }
      print '  ' * @nest, '</dl>', "\n"
    end

  end

  class Hn
    attr :nsect, true
    attr :child, true

    def markup
      n = nsect.split('.').size
      print  "<h#{n}>"
      child.markup
      print "</h#{n}>"
    end

  end

end

class Page
  def initialize
    @item = []
  end

  def push(item)
    @item.push item
  end

  def write
    @item.each { |i| i.markup }
  end

end

class Rbdoc
  def initialize
    @srcfile = $<
    @title  = ''
    @author = ''
    @mailto = ''
    @lang = ''
    @lev = 0
    @text = "".extend(Html::Plain)
    @page = Page.new
    @header = ''
    @footer = ''
    @method_switch = false
    @method_marker = '*'
    @method_indent = 4
    @mark_p = false
    @dl = []; 0.upto(16) { |i| @dl[i] = Html::Dl.new i }
    @page.push @dl[0]
    @line_env = nil
  end
  attr :title
  attr :author
  attr :mailto
  attr :lang
  attr :header
  attr :footer

  def add_item(item,lev=@lev)
    hi, lo = lev/2, lev%2
    case lo
    when 0; @dl[hi].dt_push item
    when 1; @dl[hi].dd_push item
    end
    _hi, _lo = @lev/2, @lev%2
    case _hi <=> hi
    when -1; (_hi+1).upto(hi) { |i| @dl[i-1].push @dl[i] }
    when +1; (hi+1).upto(_hi) { |i| @dl[i] = Html::Dl.new i }
    end
    @lev = lev
  end

  def method_row?
    @lev == @method_indent/2 && @mark_p ^ @method_switch
  end

  def parse_line(line)
    $_ = line
    chomp!
    case $_
    when /^%(\w+)(:|=)/
      @line_env = :plain
      act = $1
      sym = $2
      case sym
      when ':'
        src = "".extend(Html::Plain)
        while @srcfile.gets
          break if /^%$/
          chomp!
          src << "\e" << $_
        end
        @text = Rbml::here_decode(src,act).extend(Html::Plain)
        @text.act = act
        add_item @text
        @text = "".extend(Html::Plain)
        add_item @text
      when '='
        case act
        when 'lang'
          @lang = $'
        when 'method_switch'
          @method_switch = $' == 'true'
        when 'method_marker'
          @method_marker = $'
        when 'method_indent'
          @method_indent = $'.to_i
        end
      end
    when 'header:'
      @line_env = :header
      @header = ''
      while @srcfile.gets
        break if /^end\b/
        @header << $_
      end
    when 'footer:'
      @line_env = :footer
      @footer = ''
      while @srcfile.gets
        break if /^end\b/
        @footer << $_
      end
    when /^title:/
      @line_env = :title
      @title = $'.strip
    when /^author:/
      @line_env = :author
      @author = $'.strip
    when /^mailto:/
      @line_env = :mailto
      @mailto = $'.strip
    when /^include:/
      @line_env = :include
      src = "".extend(Html::Plain)
      IO.foreach($'.strip) { |i| src << i }
      @text = src#.extend(Html::Plain)
      add_item @text
    when /^#/
    when /^=/
    when /^$/
      if method_row?
        @text.replace("%method:#@text% ")#.extend(Html::Plain)
      else
        @text << "\f" if @line_env == :plain
      end
      @text = "".extend(Html::Plain)
      add_item @text
    else
      /^( *)(.*)/ =~ $_
      indent, string = $1, $2
      lev = indent.length / 2
      Html::html_encode string
      case string
      when /^((\d\.)+) (.*)/
        @line_env = :section
        item = Html::Hn.new
        item.nsect = $1.extend(Html::Plain)
        item.child = $3.strip.extend(Html::Plain)
        add_item item,lev
      else
        @line_env = :plain
        case @lang
        when 'jp'
          string.gsub!(/-$/, ' ')
        end

        errp @method_marker if $DEBUG
        regexp = Regexp.new '^\\' << @method_marker << ' '
        mark_p = !(regexp =~ string).nil?
        if mark_p
          indent << '  '
          lev+= 1
          string = $'
        end

        if @lev == lev
          case @lang
          when 'en'
            @text << "\e" << string
          when 'jp'
            @text << string
          else
            if Wide::space? @text, string
              @text << "\e" << string
            else
              @text << string
            end
          end
        else
          if method_row? && ! @text.empty?
            @text.replace("%method:#@text% ")#.extend(Html::Plain)
          end
          @text = string.extend(Html::Plain)
          add_item @text,lev
        end
        @mark_p = mark_p
      end
    end
  end
  
  def parse
    while @srcfile.gets
      parse_line $_
    end
  end
  
  def src=(fname)
    @srcname = fname.tr('\\', '/')
    @srcfile = open fname, 'r'
  end
  
  def out=(fname)
    @outname = fname.tr('\\', '/')
    $> = open fname, 'w'
  end
  
  def nav_apply
    return if @srcname.nil?

    dirname = File.dirname(@srcname)
    navname = dirname + '/nav.rb'
    return if ! test ?e, navname
    load navname

    tree_fn = dirname + '/nav.dat'
    tree = IO.readlines tree_fn

    tree_lev = []
    tree_cap = []
    tree_cap_up = []
    cnt = 0

    tree.each { |i|
      i.chomp!
      /^ */ =~ i
      indent, string = $&, $'
      cap, cap_up = string.split
      lev = indent.length / 2
      tree_lev[cnt] = lev
      tree_cap[cnt] = cap
      tree_cap_up[cnt] = cap_up
      cnt+= 1
    }

    basename = File.basename(@srcname, '.txt')
    cur = tree_cap.rindex(basename)

    if cur.nil?
      $stderr.print "#{basename} is not found in nav.dat\n"
      cur = 0
    end

    begin
      cnt = cur
      begin cnt-= 1 end while tree_lev[cnt] >= tree_lev[cur]
    rescue
      cnt+= 1
      cnt = cur if tree_lev[cnt] >= tree_lev[cur]
    end if tree_cap_up[cur].nil?

    top  = if tree_cap.empty?
      'top'
    else
      Html::html_anchor tree_cap[0], 'top'
    end
    up   =
      Html::html_anchor tree_cap_up[cur] || tree_cap[cnt], 'up'
    prev = if cur == 0
      'prev'
    else
      Html::html_anchor tree_cap[cur-1], 'prev'
    end
    ncxt = if cur == tree_cap.size-1
      'next'
    else 
      Html::html_anchor tree_cap[cur+1], 'next'
    end

    Nav.set_tree top, up, prev, ncxt

    @header << Nav.header
    @footer << Nav.footer
  end

  def write
  begin
    nav_apply
  rescue
    p $!
  end
    parse
    print <<EOT
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<html lang="ja">
<head>
<title>#{title}</title>
<link rev=made href="#{mailto}">
<link rel=stylesheet type="text/css" href="ruby.css">
<meta http-equiv="Content-Type" content="text/html; charset=shift_jis">
<meta name="Author" content="#{author}">
</head>

<body>
#{header}
EOT
    @page.write
    print <<EOT
#{footer}
<div align="right">
author: <a href="mailto:#{mailto}">#{author}</a>
</div>

</body>
</html>
EOT
  end
end
